;============================================================================
;                    Inertia Realtime 3D Rendering Engine
;         Copyright (c) 1996,1997 Alex Chalfin. All Rights Reserved.
;          DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
;============================================================================

;===================
; PLYUNPTT.ASM
; code: Alex Chalfin
;===================

; Transparent Unshaded Texture (perspective) polygon filler
;
;================================================================
; Polygon specific defines
;================================================================
;
__INTERP_EXTRA__  EQU 0        ; must be defined for everything but flat
;__INTERP_C__      EQU 0        ; must be defined for goruaud shading
__INTERP_UV__     EQU 0        ; must be defined for texture mapping
;__INTERP_ENV__    EQU 0        ; must be defined for env mapping (phong)
;

.486p                           ; 486 protected mode
.Model Flat,C                   ; with flat memory model

DATA SEGMENT PARA PUBLIC USE32  ;
                                ;
public TransUnshadedPerspTexturePolygonASM_;
                                ;
include strucdef.asm            ; Inertia structure references
include varsasm.inc             ; external variable references
                                ;
DATA ENDS                       ;
                                ;
;================================================================
; Code Segment
;================================================================
                                ;
CODE SEGMENT PARA PUBLIC USE32  ;
                                ;
;===========================================================================
; Macros
;===========================================================================
                                ;
;================================================================
;  AFFINE_LOOP
;================================================================
                                ; on entry, edi points to the correct
                                ; screen location to start drawing.
AFFINE_LOOP Macro               ;
Local @@FillLoop                ;
                                ;
  Mov  eax,[LAffU]              ;
  Mov  ebx,[LAffV]              ;
                                ;
  Sar  eax,8                    ;
  Sar  ebx,8                    ;
                                ;
  Mov  edx,[AffDU]              ;
  Mov  esi,[AffDV]              ;
                                ;
  Sar  edx,8                    ;
  Sar  esi,8                    ;
                                ;
  Mov  [HDU8],edx               ;
  Mov  [HDV8],esi               ;
                                ;
  Mov  esi,[DrawWidth]          ;
                                ;
  Xor  edx,edx                  ;
  Xor  ecx,ecx                  ;
                                ; add in the intensity offset
                                ; On entering loop:
                                ; edi = screen position
                                ; eax = 8.8 texture U value
                                ; ebx = 8.8 texture V value
  Push  ebp                     ;
 @@FillLoop:                    ;
  Mov  dl,ah                    ;
  Mov  dh,bh                    ;
                                ;
  Mov  ebp,[TextureOfsTable]    ;
                                ;
  Add  eax,[HDU8]               ;
  Add  ebx,[HDV8]               ;
                                ;
  Mov  dl,[ebp+edx]             ; get texture color
  Mov  dh,[edi]                 ;
                                ;
  Mov  ebp,[TransparencyTable]  ;
  Inc  edi                      ;
                                ;
  Mov  cl,[ebp+edx]             ;
                                ;
  Dec  esi                      ;
  Mov  [edi-1],cl               ;
                                ;
  Jnz @@FillLoop                ;
  Pop  ebp                      ;
                                ;
EndM                            ;
                                ;
;================================================================
;  PERSP_LOOP
;================================================================
                                ; on entry, edi points to the correct
                                ; screen location to start drawing.
                                ; No registers need to be saved.
                                ; the FPU stack should be clear after execution
PERSP_LOOP Macro                ;
Local @@SubdivisionWhile,@@ExitSubWhile

 @@SubdivisionWhile:            ;
  Mov  eax,[SpanCount]          ;
  Cmp  eax,0                    ;
  Jle  @@ExitSubWhile           ;
                                ;
  Mov  ebx,eax                  ;
  Sub  eax,PERSPSPAN            ;
                                ;
  Mov  [SpanCount],eax          ;
                                ;
;----------------------------------------------------------------------
; This piece of code is MIN(SpanCount, PERSPSPAN) without branching.
; Branching can have a *huge* penalty on ppro machines.
;----------------------------------------------------------------------
                                ;
  Mov  eax,ebx                  ; copy the current span width
  Mov  ebx,PERSPSPAN            ; the length of a affine span
  Cmp  ebx,eax                  ; compare the two
  Sbb  ecx,ecx                  ; if (ebx < eax), ecx = 0xffffffff, else ecx = 0
  Mov  edx,ecx                  ; copy the flag over
  Not  ecx                      ; take the bitwise inverse
  And  eax,ecx                  ; mask the origional value
  And  ebx,edx                  ; mask the max value
  Or   eax,ebx                  ; store the value
                                ;
  Mov  [DrawWidth],eax          ;
                                ;
;-----------------------------------------------------------------------
                                ;
  fild  dword ptr [DrawWidth]   ;
  fld   st(0)                   ;
                                ;
  fmul  dword ptr [HorizontalDU];
  fadd  dword ptr [hpu]         ;
  fstp  dword ptr [hpu]         ;
                                ;
  fld   st(0)                   ;
                                ;
  fmul  dword ptr [HorizontalDV];
  fadd  dword ptr [hpv]         ;
  fstp  dword ptr [hpv]         ;
                                ;
  fmul  dword ptr [HorizontalDZ];
  fadd  dword ptr [hpz]         ;
  fst   dword ptr [hpz]         ; store without pop to keep hpz value on the
                                ; FPU stack
;---------------------------------------------------------------------------
                                ;
  fld1                          ; load 1.0
  fdivrp st(1)                  ; 1.0 / hpz
                                ;
  fld    st(0)                  ; store 1.0/hpz again
  fmul   dword ptr [hpu]        ;
  fmul   dword ptr [_tmp_fixone];
                                ;
  fist   dword ptr [RAffU]      ; store RAffU
  fisub  dword ptr [LAffU]      ; calculate (RAffU - LAffU) in 16.16 fixed
                                ;
  fmul   dword ptr [DivTable+eax*4];  eax = DrawWidth
                                ;
  fistp  dword ptr [AffDU]      ; affine delta U in 16.16 fixed point
                                ;
                                ;
  fmul   dword ptr [hpv]        ;
  fmul   dword ptr [_tmp_fixone];
                                ;
  fist   dword ptr [RAffV]      ; store RAffU
  fisub  dword ptr [LAffV]      ; calculate (RAffU - LAffU) in 16.16 fixed
                                ;
  fmul   dword ptr [DivTable+eax*4];  eax = DrawWidth
                                ;
  fistp  dword ptr [AffDV]      ; affine delta U in 16.16 fixed point
                                ;
;-----------------------------------------------------------------------
; Draw the affine pixels. LAffU,LAffV are the starting texel
; values in 16.16 fixed and AffDU,AffDV are the deltas
;-----------------------------------------------------------------------
                                ;
  AFFINE_LOOP                   ;
                                ;
;-----------------------------------------------------------------------
                                ;
  Mov    eax,[RAffU]            ;
  Mov    ebx,[RAffV]            ;
                                ;
  Mov    [LAffU],eax            ;
  Mov    [LAffV],ebx            ;
                                ;
  Jmp  @@SubdivisionWhile       ;
                                ;
 @@ExitSubWhile:                ;
                                ;
EndM                            ;
                                ;
;================================================================
;  INNER_LOOP
;================================================================
                                ; All registers are available for loop
                                ; macro. Nothing needs to be saved/restored
                                ; it is all handled in PERSPASM.INC
                                ;
INNER_LOOP Macro  Left, Right, Height, Side;
Local  @@HeightLooper,@@ExitHeight ; Local labels
                                ;
 @@HeightLooper:                ;
  Mov  eax,[Height]             ;
  Cmp  eax,0                    ;
  Jle @@ExitHeight              ;
                                ;
  Dec  [Height]                 ; decrement the height value by 1
                                ;
  Mov  ebx,[Left&X]             ;
  Add  ebx,65535                ;
  And  ebx,0ffff0000h           ;
                                ;
  Mov  edx,ebx                  ;
  Sub  edx,[Left&X]             ; edx = (ceil(LeftX) - LeftX) in 16.16 fixed
                                ;
  Mov  [_tmp_flt],edx           ;
                                ;
  fild dword ptr [_tmp_flt]     ;
  fmul dword ptr [_tmp_fxd2flt] ; convert the fixed value to a float
                                ;
  fld  st(0)                    ; load the value again
;--------------------------------
; Sub-pixel correction
;--------------------------------
                                ;
  fmul dword ptr [HorizontalDU] ; perform sub-pixel correction
  fadd dword ptr [Side&U]       ; add in starting U/Z value
                                ;
  fstp dword ptr [hpu]          ; store the horizontal perspective u value
                                ;-------------------------------------------
  fld  st(0)                    ;
  fmul dword ptr [HorizontalDV] ; perform sub-pixel correction
  fadd dword ptr [Side&V]       ; add in starting V/Z value
                                ;
  fstp dword ptr [hpv]          ; store the horizontal perspective v value
                                ;-------------------------------------------
  fmul dword ptr [HorizontalDZ] ; perform sub-pixel correction
  fadd dword ptr [Side&Z]       ; add in starting 1/Z value
                                ;
  fst  dword ptr [hpz]          ; store the horizontal perspective z value
;-------------------------------------------------------------------------
                                ;
  fld1                          ; load 1.0
  fdivrp st(1)                  ; calculate 1/hpz for the starting loc
                                ;
  fld    st(0)                  ; reload the 1/z value
                                ;
  fmul   dword ptr [hpu]        ; get the perspective correct U coordiante.
  fmul   dword ptr [_tmp_fixone]; convert to fixed point
                                ;
  fistp  dword ptr [LAffU]      ; store the left affine U value
                                ;
  fmul   dword ptr [hpv]        ; get the perspective correct V coordiante.
  fmul   dword ptr [_tmp_fixone]; convert to fixed point
                                ;
  fistp  dword ptr [LAffV]      ; store the left affine V value
;--------------------------------------------------------------------------
                                ;
  Mov    eax,[Left&X]           ;
  Mov    ebx,[Right&X]          ;
                                ;
  Add    eax,65535              ;
  Dec    ebx                    ;
                                ;
  Sar    eax,16                 ;
  Sar    ebx,16                 ;
                                ;
  Sub    ebx,eax                ;
  Inc    ebx                    ;
                                ;
  Mov    edi,eax                ;
  Add    edi,[YAddress]         ; edi points to the screen location
                                ;
  Mov    [SpanCount],ebx        ; the length of the span
;--------------------------------------------------------------------------
                                ;
  PERSP_LOOP                    ; perspective span loop
                                ;
;--------------------------------------------------------------------------
                                ;
  Mov  edx,[YAddress]           ; go to the next scanline
  Add  edx,[LineSize]           ;
  Mov  [YAddress],edx           ; YAddress += LineSize
                                ;
  fld  dword ptr [Side&U]       ; increment the U span
  fadd dword ptr [Side&DU]      ;
  fstp dword ptr [Side&U]       ;
                                ;
  fld  dword ptr [Side&V]       ; increment the V span
  fadd dword ptr [Side&DV]      ;
  fstp dword ptr [Side&V]       ;
                                ;
  fld  dword ptr [Side&Z]       ; increment the Z span
  fadd dword ptr [Side&DZ]      ;
  fstp dword ptr [Side&Z]       ;
                                ;
  Mov  eax,[Left&X]             ;
  Mov  ebx,[Right&X]            ;
  Add  eax,[Left&DX]            ;
  Add  ebx,[Right&DX]           ;
  Mov  [Left&X],eax             ; LeftX += LeftDX
  Mov  [Right&X],ebx            ; RightX += RightDX
                                ;
  Jmp @@HeightLooper            ;
                                ;
 @@ExitHeight:                  ;
EndM                            ;
                                ;
;===========================================================================
; UnshadedPerspTexturePolygonASM
;===========================================================================
TransUnshadedPerspTexturePolygonASM_ Proc
                                ;
  Mov [VirtualAddress],esi      ;
  Mov [LineSize],eax            ;
                                ;
  Mov   eax,[edi].Texture       ;
  Mov   [TextureOfsTable],eax   ;
                                ;
include perspasm.inc            ; include perspective interpolation
                                ;
  Ret                           ;
                                ;
TransUnshadedPerspTexturePolygonASM_ EndP  ;


CODE ENDS
END


