;
;                    Inertia Realtime 3D Rendering Engine
;    Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
;          DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
;

;
; PLYFLRE.ASM
; code: Alex Chalfin
;

; Unshaded Reflection (affine) polygon filler
;
;================================================================
; Polygon specific defines
;================================================================
;
__INTERP_EXTRA__  EQU 0        ; must be defined for everything but flat
;__INTERP_C__      EQU 0        ; must be defined for goruaud shading
;__INTERP_UV__     EQU 0        ; must be defined for texture mapping
__INTERP_ENV__    EQU 0        ; must be defined for env mapping (phong)
;

.486p                           ; 486 protected mode
.Model Flat,C                   ; with flat memory model

DATA SEGMENT PARA PUBLIC USE32  ;
                                ;
public UnshadedReflectionPolygonASM_;
                                ;
include strucdef.asm            ;  Inertia structure references
include varsasm.inc             ; external variable references
                                ;
DATA ENDS                       ;
                                ;
;================================================================
; Code Segment
;================================================================
                                ;
CODE SEGMENT PARA PUBLIC USE32  ;
                                ;
;===========================================================================
; Macros
;===========================================================================
                                ;
                                ; All registers are available for loop
                                ; macro. Nothing needs to be saved/restored
                                ; it is all handled in AFFASM.INC
                                ;
INNER_LOOP Macro  Left, Right, Height, Side;
Local  @@HeightLooper,@@ExitHeight,@@SkipFill,@@FillLoop ; Local labels
                                ;
 @@HeightLooper:                ;
  Mov  eax,[Height]             ;
  Cmp  eax,0                    ;
  Jle @@ExitHeight              ;
                                ;
  Dec  [Height]                 ;  decrement the height value by 1
                                ;
  Mov  ebx,[Left&X]             ;
  Add  ebx,65535                ;
  And  ebx,0ffff0000h           ;
                                ;
  Mov  edx,ebx                  ;
  Sub  edx,[Left&X]             ;
                                ; edx contains sub-pixel correction factor
  Mov  esi,edx                  ;
                                ;
  Mov  eax,[HorizontalDS]       ;
  IMul edx                      ;
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&S]             ; eax is the starting horizontal color
                                ;
  Mov  [SStart],eax             ;
                                ;
  Mov  edx,esi                  ; get sup-pixel correction factor
  Mov  eax,[HorizontalDT]       ;
  IMul edx                      ;
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&T]             ; eax is the starting horizontal color
                                ;
  Mov  [TStart],eax             ;
                                ;
  Sar  ebx,16                   ;
                                ;
  Mov  ecx,[Right&X]            ;
  Dec  ecx                      ;
  Sar  ecx,16                   ;
                                ;
  Sub  ecx,ebx                  ;
  Inc  ecx                      ;
  Cmp  ecx,0                    ;
  Jle @@SkipFill                ;
                                ;
  Mov  edi,[YAddress]           ;
  Add  edi,ebx                  ;
                                ;
  Mov  eax,[SStart]             ;
  Mov  ebx,[TStart]             ;
                                ;
  Sar  eax,8                    ;
  Sar  ebx,8                    ;
                                ;
  Mov  esi,[Color]              ;
  Mov  edx,[HorizontalDS]       ;
  Mov  esi,[HorizontalDT]       ;
                                ;
  Sar  edx,8                    ;
  Sar  esi,8                    ;
                                ;
  Mov  [HDS8],edx               ;
  Mov  [HDT8],esi               ;
  Mov  esi,ecx                  ;
                                ;
  Xor  edx,edx                  ;
  Xor  ecx,ecx                  ;
                                ; add in the intensity offset
                                ; On entering loop:
                                ; edi = screen position
                                ; eax = 8.8 texture U value
                                ; ebx = 8.8 texture V value
  Push  ebp                     ;
  Mov  ebp,[TextureOfsTable]    ;
 @@FillLoop:                    ;
  Mov  dl,ah                    ;
  Mov  dh,bh                    ;
                                ;
  Add  eax,[HDS8]               ;
  Add  ebx,[HDT8]               ;
                                ;
  Mov  dl,[ebp+edx]             ; get texture color
  Inc  edi                      ;
                                ;
  Dec  esi                      ;
  Mov  [edi-1],dl               ;
                                ;
  Jnz @@FillLoop                ;
  Pop  ebp                      ;
                                ;
 @@SkipFill:                    ;
                                ;
  Mov  edx,[YAddress]           ; go to the next scanline
  Add  edx,[LineSize]           ;
  Mov  [YAddress],edx           ; YAddress += LineSize
                                ;
  Mov  eax,[Side&S]             ; increment the U span
  Add  eax,[Side&DS]            ;
  Mov  [Side&S],eax             ;
                                ;
  Mov  eax,[Side&T]             ; increment the V span
  Add  eax,[Side&DT]            ;
  Mov  [Side&T],eax             ;
                                ;
  Mov  eax,[Left&X]             ;
  Mov  ebx,[Right&X]            ;
  Add  eax,[Left&DX]            ;
  Add  ebx,[Right&DX]           ;
  Mov  [Left&X],eax             ; LeftX += LeftDX
  Mov  [Right&X],ebx            ; RightX += RightDX
                                ;
  Jmp @@HeightLooper            ;
                                ;
 @@ExitHeight:                  ;
EndM                            ;
                                ;
;===========================================================================
; CullPolys
;===========================================================================
UnshadedReflectionPolygonASM_ Proc
                                ;
  Mov [VirtualAddress],esi      ;
  Mov [LineSize],eax            ;
                                ;
  Mov   eax,[edi].Texture       ;
  Mov   [TextureOfsTable],eax   ;
                                ;
include affasm.inc              ; include affine interpolation
                                ;
  Ret                           ;
                                ;
UnshadedReflectionPolygonASM_ EndP;


CODE ENDS
END


