/*

                     Inertia Realtime 3D Rendering Engine
     Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED

*/

#ifndef __TYPES_H__
#define __TYPES_H__

#include "define.inc"

#ifdef  SHAREWARE
#define MAX_POLYGONS   500
#endif

#define MAXTEXTURES 128
#define CORRECTRES  32

#define MAXLIGHTS   8   /* 8 light sources */
#define XLOC        9   /* index in matrix for X location */
#define YLOC        10  /* index in matrix for Y location */
#define ZLOC        11  /* index in matrix for Z location */

#define I_OK        0
#define I_MEMERR    1
#define I_FILEERR   2
#define I_FORMATERR 3

#define AMBIENT       0x00
#define LAMBERT       0x01
#define GOURAUD       0x02
#define PHONG         0x03
#define UNSHADED      0x04

#define SMOOTH        0x00
#define TEXTURE       0x10
#define REFLECTION    0x20
#define PERSPTEXTURE  0x30

#define TRANSPARENT   0x80

#define SHADE_MASK    0x0f
#define SURFACE_MASK  0x70

#define CLIP_Z        0x01
#define CLIP_LEFT     0x02
#define CLIP_RIGHT    0x04
#define CLIP_TOP      0x08
#define CLIP_BOTTOM   0x10
#define CLIP_GOURAUD  0x20
#define CLIP_TEXTURE  0x40
#define CLIP_ENV      0x80

#define CLIP_SHADEMASK  (CLIP_GOURAUD | CLIP_TEXTURE | CLIP_ENV)

typedef struct
{
  char r, g, b;
} RGBType;

typedef RGBType Palette[256];
typedef float Matrix4x3[12];

typedef struct
{
  char    *Texture;
  Palette Pal;
} TextureRec;

typedef struct
{
  float x, y, z;
} Vertex3d;

typedef struct
{
  float x, y;
} Vertex2d;

typedef struct
{
  Vertex3d Loc;
  Vertex3d Cam;
  Vertex2d Scr;
  Vertex2d Env;
  Vertex3d N;
  float    OneOverZ;
  float    Shade;
  int      Visible;
} Vertex;

typedef struct
{
  Vertex   *A, *B, *C;
  Vertex3d N;
  Vertex2d UV1, UV2, UV3;
  char    *Texture;
  char    *IBuffer;
  float    Shade;
  int      Visible;
  int      Color;
} Triangle;

typedef struct _tSortRec
{
  int              ZMid;
  int              ShadeSurface;
  struct _tSortRec *Next;
  Triangle         *Poly;
} SortRec;

typedef struct
{
  int          Allocated;
  SortRec    *SortList;
  int          Count;
  SortRec    *MSBList[256];
  SortRec    *LSBList[256];
} tRenderList;

typedef struct
{
  int NumLights;
  Vertex3d Light[MAXLIGHTS];
  Vertex3d TempLight[MAXLIGHTS];
} tLightList;

typedef struct
{
  int      NumVert;
  int      NumPoly;
  Vertex   *Vertexs;
  Triangle *Polygons;

  int      RenderMask;
  float    Radius;
  int      CorrectCount;
  int      ClipMask;
  int      Cull;
  Matrix4x3 Orientation;
} VectorObject;

typedef struct
{
  float CenterX;
  float CenterY;

  int WinMinX, WinMinY;
  int WinMaxX, WinMaxY;

  float InvCamZFac;
  float InvCamYFac;
  float CamZFac;
  float CamYFac;

  float FocalLength;
  float FieldOfView;

  float FustrumMinZ;
  float FustrumMaxZ;

  Matrix4x3  Camera;
} ViewObject;

typedef void (*TriangleFiller)(Triangle *P, char *VPage, int x, int y);
typedef void (*ShaderFunction)(ViewObject *View, VectorObject *Obj, float *Matrix);

#endif
