/*

                       Inerita Realtime 3D Rendering Engine
            Copyright (c) 1997, Alex Chalfin. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED

*/
#include "xform.h"

void SingleProject(ViewObject *View, Vertex *V)
{
  float YScale, XScale;
  float WinX = View->WinMaxX - View->WinMinX + 1;

  XScale  = (WinX / 2.0);
  YScale  = XScale * AspectRatio * View->InvCamYFac;

  V->OneOverZ = 1.0 / V->Cam.z;
  V->Scr.x = (View->CenterX + (XScale * V->Cam.x * V->OneOverZ));
  V->Scr.y = (View->CenterY - (YScale * V->Cam.y * V->OneOverZ));
}

void XForm(ViewObject *View, VectorObject *Obj, float *M)
{
  int    i;
  Vertex *V;

  for (i = 0, V = Obj->Vertexs; i < Obj->NumVert; i++, V++)
  {
    if (V->Visible == 1)
    {
      V->Cam.x = M[0] * V->Loc.x + M[3] * V->Loc.y + M[6] * V->Loc.z + M[XLOC];
      V->Cam.y = M[1] * V->Loc.x + M[4] * V->Loc.y + M[7] * V->Loc.z + M[YLOC];
      V->Cam.z = M[2] * V->Loc.x + M[5] * V->Loc.y + M[8] * V->Loc.z + M[ZLOC];
    }
  }
}

void Project(ViewObject *View, VectorObject *Obj)
{
  int      i;
  Vertex   *V;
  float    XScale, YScale;
  float    WinX = View->WinMaxX - View->WinMinX + 1;

  XScale  = (WinX / 2);
  YScale  = XScale * AspectRatio * View->InvCamYFac;

  for (i = 0, V = Obj->Vertexs; i < Obj->NumVert; i++, V++)
  {
    if (V->Visible == 1)
    {
      V->OneOverZ = 1.0 / V->Cam.z;
      V->Scr.x = (View->CenterX + (XScale * V->Cam.x * V->OneOverZ));
      V->Scr.y = (View->CenterY - (YScale * V->Cam.y * V->OneOverZ));
    }
  }
}

void InvXForm(Vertex3d *V, float *M)
{
  float   tx = V->x;
  float   ty = V->y;
  float   tz = V->z;

  V->x = M[0] * tx + M[1] * ty + M[2] * tz;
  V->y = M[3] * tx + M[4] * ty + M[5] * tz;
  V->z = M[6] * tx + M[7] * ty + M[8] * tz;
}

void SingleRotate(Vertex3d *Vector, float xa, float ya, float za)
{
  float tx, ty, tz;
  float sx, sy, sz;
  float cx, cy, cz;

  sx = sin(xa);   sy = sin(ya);   sz = sin(za);
  cx = cos(xa);   cy = cos(ya);   cz = cos(za);

  tx = (Vector->x * cz) - (Vector->y * sz);
  ty = (Vector->x * sz) + (Vector->y * cz);
  tz = Vector->z;

  Vector->x = (tx * cy) + (tz * sy);
  Vector->y = ty;
  Vector->z = (tz * cy) - (tx * sy);

  tx = Vector->x;
  ty = (Vector->y * cx) - (Vector->z * sx);
  tz = (Vector->y * sx) + (Vector->z * cx);

  Vector->x = tx;
  Vector->y = ty;
  Vector->z = tz;
}


